var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import Point from "@mapbox/point-geometry";
import { dark } from "../default_style/dark";
import { light } from "../default_style/light";
import { labelRules, paintRules } from "../default_style/style";
import { Labelers } from "../labeler";
import { painter } from "../painter";
import { sourcesToViews } from "../view";
import { xray_rules } from "../xray";
const timer = (duration) => {
    return new Promise((resolve, reject) => {
        setTimeout(() => {
            resolve();
        }, duration);
    });
};
const reflect = (promise) => {
    return promise.then((v) => {
        return { status: "fulfilled", value: v };
    }, (error) => {
        return { status: "rejected", reason: error };
    });
};
const leafletLayer = (options = {}) => {
    class LeafletLayer extends L.GridLayer {
        constructor(options = {}) {
            if (options.noWrap && !options.bounds)
                options.bounds = [
                    [-90, -180],
                    [90, 180],
                ];
            if (options.attribution == null)
                options.attribution =
                    '<a href="https://protomaps.com">Protomaps</a> © <a href="https://openstreetmap.org/copyright">OpenStreetMap</a>';
            super(options);
            const theme = options.dark ? dark : light;
            this.paint_rules =
                options.paint_rules || paintRules(theme, options.shade);
            this.label_rules =
                options.label_rules ||
                    labelRules(theme, options.shade, options.language1, options.language2);
            this.backgroundColor = options.backgroundColor;
            this.lastRequestedZ = undefined;
            this.xray = options.xray;
            this.tasks = options.tasks || [];
            this.views = sourcesToViews(options);
            this.debug = options.debug;
            const scratch = document.createElement("canvas").getContext("2d");
            this.scratch = scratch;
            this.onTilesInvalidated = (tiles) => {
                for (const t of tiles) {
                    this.rerenderTile(t);
                }
            };
            this.labelers = new Labelers(this.scratch, this.label_rules, 16, this.onTilesInvalidated);
            this.tile_size = 256 * window.devicePixelRatio;
            this.tileDelay = options.tileDelay || 3;
            this.lang = options.lang;
            // bound instance of function
            this.inspector = this.inspect(this);
        }
        setDefaultStyle(darkOption, shade, language1, language2) {
            const theme = darkOption ? dark : light;
            this.paint_rules = paintRules(theme, shade);
            this.label_rules = labelRules(theme, shade, language1, language2);
        }
        renderTile(coords, element, key, done = () => { }) {
            return __awaiter(this, void 0, void 0, function* () {
                this.lastRequestedZ = coords.z;
                const promises = [];
                for (const [k, v] of this.views) {
                    const promise = v.getDisplayTile(coords);
                    promises.push({ key: k, promise: promise });
                }
                const tile_responses = yield Promise.all(promises.map((o) => {
                    return o.promise.then((v) => {
                        return { status: "fulfilled", value: v, key: o.key };
                    }, (error) => {
                        return { status: "rejected", reason: error, key: o.key };
                    });
                }));
                const prepared_tilemap = new Map();
                for (const tile_response of tile_responses) {
                    if (tile_response.status === "fulfilled") {
                        prepared_tilemap.set(tile_response.key, [tile_response.value]);
                    }
                    else {
                        if (tile_response.reason.name === "AbortError") {
                            // do nothing
                        }
                        else {
                            console.error(tile_response.reason);
                        }
                    }
                }
                if (element.key !== key)
                    return;
                if (this.lastRequestedZ !== coords.z)
                    return;
                yield Promise.all(this.tasks.map(reflect));
                if (element.key !== key)
                    return;
                if (this.lastRequestedZ !== coords.z)
                    return;
                const layout_time = this.labelers.add(coords.z, prepared_tilemap);
                if (element.key !== key)
                    return;
                if (this.lastRequestedZ !== coords.z)
                    return;
                const label_data = this.labelers.getIndex(coords.z);
                if (!this._map)
                    return; // the layer has been removed from the map
                const center = this._map.getCenter().wrap();
                const pixelBounds = this._getTiledPixelBounds(center);
                const tileRange = this._pxBoundsToTileRange(pixelBounds);
                const tileCenter = tileRange.getCenter();
                const priority = coords.distanceTo(tileCenter) * this.tileDelay;
                yield timer(priority);
                if (element.key !== key)
                    return;
                if (this.lastRequestedZ !== coords.z)
                    return;
                const BUF = 16;
                const bbox = {
                    minX: 256 * coords.x - BUF,
                    minY: 256 * coords.y - BUF,
                    maxX: 256 * (coords.x + 1) + BUF,
                    maxY: 256 * (coords.y + 1) + BUF,
                };
                const origin = new Point(256 * coords.x, 256 * coords.y);
                element.width = this.tile_size;
                element.height = this.tile_size;
                const ctx = element.getContext("2d");
                ctx.setTransform(this.tile_size / 256, 0, 0, this.tile_size / 256, 0, 0);
                ctx.clearRect(0, 0, 256, 256);
                if (this.backgroundColor) {
                    ctx.save();
                    ctx.fillStyle = this.backgroundColor;
                    ctx.fillRect(0, 0, 256, 256);
                    ctx.restore();
                }
                let painting_time = 0;
                let paint_rules = this.paint_rules;
                if (this.xray) {
                    paint_rules = xray_rules(prepared_tilemap, this.xray);
                }
                painting_time = painter(ctx, coords.z, prepared_tilemap, this.xray ? null : label_data, paint_rules, bbox, origin, false, this.debug);
                if (this.debug) {
                    ctx.save();
                    ctx.fillStyle = this.debug;
                    ctx.font = "600 12px sans-serif";
                    ctx.fillText(`${coords.z} ${coords.x} ${coords.y}`, 4, 14);
                    ctx.font = "12px sans-serif";
                    let ypos = 28;
                    for (const [k, v] of prepared_tilemap) {
                        const dt = v[0].data_tile;
                        ctx.fillText(`${k + (k ? " " : "") + dt.z} ${dt.x} ${dt.y}`, 4, ypos);
                        ypos += 14;
                    }
                    ctx.font = "600 10px sans-serif";
                    if (painting_time > 8) {
                        ctx.fillText(`${painting_time.toFixed()} ms paint`, 4, ypos);
                        ypos += 14;
                    }
                    if (layout_time > 8) {
                        ctx.fillText(`${layout_time.toFixed()} ms layout`, 4, ypos);
                    }
                    ctx.strokeStyle = this.debug;
                    ctx.lineWidth = 0.5;
                    ctx.beginPath();
                    ctx.moveTo(0, 0);
                    ctx.lineTo(0, 256);
                    ctx.stroke();
                    ctx.lineWidth = 0.5;
                    ctx.beginPath();
                    ctx.moveTo(0, 0);
                    ctx.lineTo(256, 0);
                    ctx.stroke();
                    ctx.restore();
                }
                done();
            });
        }
        rerenderTile(key) {
            for (const unwrapped_k in this._tiles) {
                const wrapped_coord = this._wrapCoords(this._keyToTileCoords(unwrapped_k));
                if (key === this._tileCoordsToKey(wrapped_coord)) {
                    this.renderTile(wrapped_coord, this._tiles[unwrapped_k].el, key);
                }
            }
        }
        clearLayout() {
            this.labelers = new Labelers(this.scratch, this.label_rules, 16, this.onTilesInvalidated);
        }
        rerenderTiles() {
            for (const unwrapped_k in this._tiles) {
                const wrapped_coord = this._wrapCoords(this._keyToTileCoords(unwrapped_k));
                const key = this._tileCoordsToKey(wrapped_coord);
                this.renderTile(wrapped_coord, this._tiles[unwrapped_k].el, key);
            }
        }
        createTile(coords, showTile) {
            const element = L.DomUtil.create("canvas", "leaflet-tile");
            element.lang = this.lang;
            const key = this._tileCoordsToKey(coords);
            element.key = key;
            this.renderTile(coords, element, key, () => {
                showTile(null, element);
            });
            return element;
        }
        _removeTile(key) {
            const tile = this._tiles[key];
            if (!tile) {
                return;
            }
            tile.el.removed = true;
            tile.el.key = undefined;
            L.DomUtil.removeClass(tile.el, "leaflet-tile-loaded");
            tile.el.width = tile.el.height = 0;
            L.DomUtil.remove(tile.el);
            delete this._tiles[key];
            this.fire("tileunload", {
                tile: tile.el,
                coords: this._keyToTileCoords(key),
            });
        }
        queryFeatures(lng, lat) {
            const featuresBySourceName = new Map();
            for (const [sourceName, view] of this.views) {
                featuresBySourceName.set(sourceName, view.queryFeatures(lng, lat, this._map.getZoom()));
            }
            return featuresBySourceName;
        }
        inspect(layer) {
            return (ev) => {
                const typeGlyphs = ["◎", "⟍", "◻"];
                const wrapped = layer._map.wrapLatLng(ev.latlng);
                const resultsBySourceName = layer.queryFeatures(wrapped.lng, wrapped.lat);
                let content = "";
                let firstRow = true;
                for (const [sourceName, results] of resultsBySourceName) {
                    for (const result of results) {
                        if (this.xray && this.xray !== true) {
                            if (!((this.xray.dataSource || "") === sourceName &&
                                this.xray.dataLayer === result.layerName)) {
                                continue;
                            }
                        }
                        content = `${content}<div style="margin-top:${firstRow ? 0 : 0.5}em">${typeGlyphs[result.feature.geomType - 1]} <b>${sourceName} ${sourceName ? "/" : ""} ${result.layerName}</b> ${result.feature.id || ""}</div>`;
                        for (const prop in result.feature.props) {
                            content = `${content}<div style="font-size:0.9em">${prop} = ${result.feature.props[prop]}</div>`;
                        }
                        firstRow = false;
                    }
                }
                if (firstRow) {
                    content = "No features.";
                }
                L.popup()
                    .setLatLng(ev.latlng)
                    .setContent(`<div style="max-height:400px;overflow-y:scroll;padding-right:8px">${content}</div>`)
                    .openOn(layer._map);
            };
        }
        addInspector(map) {
            return map.on("click", this.inspector);
        }
        removeInspector(map) {
            return map.off("click", this.inspector);
        }
    }
    return new LeafletLayer(options);
};
export { leafletLayer };
